# ruff: noqa: UP006 UP007
"""
TODO:
 - SocketClientAddress family / tuple pairs
  + codification of https://docs.python.org/3/library/socket.html#socket-families
"""
import abc
import dataclasses as dc
import socket
import typing as ta


SocketAddress = ta.Any


SocketHandlerFactory = ta.Callable[[SocketAddress, ta.BinaryIO, ta.BinaryIO], 'SocketHandler']


##


@dc.dataclass(frozen=True)
class SocketAddressInfoArgs:
    host: ta.Optional[str]
    port: ta.Union[str, int, None]
    family: socket.AddressFamily = socket.AddressFamily.AF_UNSPEC
    type: int = 0
    proto: int = 0
    flags: socket.AddressInfo = socket.AddressInfo(0)


@dc.dataclass(frozen=True)
class SocketAddressInfo:
    family: socket.AddressFamily
    type: int
    proto: int
    canonname: ta.Optional[str]
    sockaddr: SocketAddress


def get_best_socket_family(
        host: ta.Optional[str],
        port: ta.Union[str, int, None],
        family: ta.Union[int, socket.AddressFamily] = socket.AddressFamily.AF_UNSPEC,
) -> ta.Tuple[socket.AddressFamily, SocketAddress]:
    """https://github.com/python/cpython/commit/f289084c83190cc72db4a70c58f007ec62e75247"""

    infos = socket.getaddrinfo(
        host,
        port,
        family,
        type=socket.SOCK_STREAM,
        flags=socket.AI_PASSIVE,
    )
    ai = SocketAddressInfo(*next(iter(infos)))
    return ai.family, ai.sockaddr


##


class SocketHandler(abc.ABC):
    def __init__(
            self,
            client_address: SocketAddress,
            rfile: ta.BinaryIO,
            wfile: ta.BinaryIO,
    ) -> None:
        super().__init__()

        self._client_address = client_address
        self._rfile = rfile
        self._wfile = wfile

    @abc.abstractmethod
    def handle(self) -> None:
        raise NotImplementedError



# ruff: noqa: N802 N803
from . import runtime as antlr4


class ParseError(Exception):
    pass


class SilentRaisingErrorListener(antlr4.ErrorListener):
    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        raise ParseError(recognizer, offendingSymbol, line, column, msg, e)



import dataclasses as dc
import json
import os.path
import plistlib
import re
import shutil
import subprocess
import sys
import typing as ta

from .. import check
from .. import lang


if ta.TYPE_CHECKING:
    docker = lang.proxy_import('..docker')
else:
    from .. import docker


##


PYCHARM_HOSTED_ENV_VAR = 'PYCHARM_HOSTED'


def is_pycharm_hosted() -> bool:
    return PYCHARM_HOSTED_ENV_VAR in os.environ


##


DARWIN_PYCHARM_HOME = '/Applications/PyCharm.app'


def read_darwin_pycharm_info_plist() -> ta.Mapping[str, ta.Any] | None:
    plist_file = os.path.join(DARWIN_PYCHARM_HOME, 'Contents', 'Info.plist')
    if not os.path.isfile(plist_file):
        return None

    with open(plist_file, 'rb') as f:
        root = plistlib.load(f)

    return root


#


UBUNTU_PYCHARM_HOME = '/snap/pycharm-professional/current'


def read_ubuntu_pycharm_product_info() -> ta.Mapping[str, ta.Any] | None:
    json_file = os.path.join(UBUNTU_PYCHARM_HOME, 'product-info.json')
    if not os.path.isfile(json_file):
        return None

    with open(json_file) as f:
        root = json.load(f)

    return root


#


@lang.cached_function
def get_pycharm_version() -> str | None:
    if sys.platform == 'darwin':
        plist = read_darwin_pycharm_info_plist()
        if plist is None:
            return None

        ver = check.non_empty_str(plist['CFBundleVersion'])
        check.state(ver.startswith('PY-'))
        return ver[3:]

    elif sys.platform == 'linux':
        if shutil.which('lsb_release') is not None:
            lsb_id = subprocess.check_output(['lsb_release', '-is']).decode().strip()
            if lsb_id == 'Ubuntu':
                pi = read_ubuntu_pycharm_product_info()
                if pi is not None:
                    ver = check.non_empty_str(pi['buildNumber'])
                    return ver

        return None

    else:
        return None


##


def _import_pydevd_pycharm(*, version: str | None = None) -> ta.Any:
    if (
            'pydevd_pycharm' in sys.modules or
            (version is None and lang.can_import('pydevd_pycharm'))
    ):
        # Can't unload, nothing we can do
        import pydevd_pycharm  # noqa
        return pydevd_pycharm

    proc = subprocess.run([  # noqa
        sys.executable,
        '-m', 'pip',
        'show',
        'pydevd_pycharm',
    ], stdout=subprocess.PIPE)

    if not proc.returncode:
        info = {
            k: v.strip()
            for l in proc.stdout.decode().splitlines()
            if (s := l.strip())
            for k, _, v in [s.partition(':')]
        }

        installed_version = info['Version']
        if installed_version == version:
            import pydevd_pycharm  # noqa
            return pydevd_pycharm

        subprocess.check_call([
            sys.executable,
            '-m', 'pip',
            'uninstall', '-y',
            'pydevd_pycharm',
        ])

    subprocess.check_call([
        sys.executable,
        '-m', 'pip',
        'install',
        'pydevd_pycharm' + (f'=={version}' if version is not None else ''),
    ])

    import pydevd_pycharm  # noqa
    return pydevd_pycharm


@dc.dataclass(frozen=True)
class PycharmRemoteDebugger:
    host: str | None
    port: int
    version: str | None = None

    def __str__(self) -> str:
        return ''.join([
            f'@{self.version}:' if self.version is not None else '',
            f'{self.host}:' if self.host is not None else '',
            str(self.port),
        ])

    @classmethod
    def parse(cls, s: str) -> 'PycharmRemoteDebugger':
        if (m := re.fullmatch(r'(@(?P<version>[^:]+):)?((?P<host>[^:]+):)?(?P<port>\d+)', s)) is None:
            raise ValueError(s)
        gd = m.groupdict()
        return cls(
            host=gd.get('host'),
            port=int(gd['port']),
            version=gd.get('version'),
        )


def pycharm_remote_debugger_attach(prd: PycharmRemoteDebugger) -> None:
    host = prd.host
    if host is None:
        if (
                sys.platform == 'linux' and
                docker.is_likely_in_docker() and
                docker.get_docker_host_platform() == 'darwin'
        ):
            host = docker.DOCKER_FOR_MAC_HOSTNAME
        else:
            host = 'localhost'

    version = prd.version
    if version is None and host in ('localhost', '127.0.0.1'):
        version = get_pycharm_version()

    if ta.TYPE_CHECKING:
        import pydevd_pycharm  # noqa
    else:
        pydevd_pycharm = _import_pydevd_pycharm(version=version)

    pydevd_pycharm.settrace(
        host,
        port=prd.port,
        stdoutToServer=True,
        stderrToServer=True,
    )



"""
TODO:
 - go-style tags: +slow,-ci, ...
"""
from _pytest.main import resolve_collection_argument  # noqa
import pytest

from ._registry import register


@register
class SkipsPlugin:

    def pytest_collection_modifyitems(self, session, items):
        dct: dict[str, set[str]] = {}
        for arg in session.config.args:
            ca = resolve_collection_argument(
                session.config.invocation_params.dir,
                arg,
                as_pypath=session.config.option.pyargs,
            )
            if ca.path.is_file():
                dct.setdefault(ca.path.as_posix(), set()).update(ca.parts)

        skip = pytest.mark.skip(reason='skipped (not specified alone)')
        for item in items:
            if 'skip_unless_alone' in item.keywords:
                if item.name not in dct.get(item.fspath.strpath, ()):
                    item.add_marker(skip)

    def pytest_configure(self, config):
        config.addinivalue_line('markers', 'skip_unless_alone: mark test as skipped unless specified alone')



# PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
# --------------------------------------------
#
# 1. This LICENSE AGREEMENT is between the Python Software Foundation ("PSF"), and the Individual or Organization
# ("Licensee") accessing and otherwise using this software ("Python") in source or binary form and its associated
# documentation.
#
# 2. Subject to the terms and conditions of this License Agreement, PSF hereby grants Licensee a nonexclusive,
# royalty-free, world-wide license to reproduce, analyze, test, perform and/or display publicly, prepare derivative
# works, distribute, and otherwise use Python alone or in any derivative version, provided, however, that PSF's License
# Agreement and PSF's notice of copyright, i.e., "Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
# 2010, 2011, 2012, 2013, 2014, 2015, 2016, 2017 Python Software Foundation; All Rights Reserved" are retained in Python
# alone or in any derivative version prepared by Licensee.
#
# 3. In the event Licensee prepares a derivative work that is based on or incorporates Python or any part thereof, and
# wants to make the derivative work available to others as provided herein, then Licensee hereby agrees to include in
# any such work a brief summary of the changes made to Python.
#
# 4. PSF is making Python available to Licensee on an "AS IS" basis.  PSF MAKES NO REPRESENTATIONS OR WARRANTIES,
# EXPRESS OR IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND DISCLAIMS ANY REPRESENTATION OR WARRANTY
# OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF PYTHON WILL NOT INFRINGE ANY THIRD PARTY
# RIGHTS.
#
# 5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL
# DAMAGES OR LOSS AS A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON, OR ANY DERIVATIVE THEREOF, EVEN IF
# ADVISED OF THE POSSIBILITY THEREOF.
#
# 6. This License Agreement will automatically terminate upon a material breach of its terms and conditions.
#
# 7. Nothing in this License Agreement shall be deemed to create any relationship of agency, partnership, or joint
# venture between PSF and Licensee.  This License Agreement does not grant permission to use PSF trademarks or trade
# name in a trademark sense to endorse or promote products or services of Licensee, or any third party.
#
# 8. By copying, installing or otherwise using Python, Licensee agrees to be bound by the terms and conditions of this
# License Agreement.
import dataclasses as dc
import functools
import os.path
import struct
import time
import typing as ta

from ... import cached
from ... import check
from ... import lang
from ..generators import BytesSteppedGenerator
from ..generators import BytesSteppedReaderGenerator
from ..generators.readers import PrependableBytesGeneratorReader
from .base import Compression
from .base import IncrementalCompression
from .codecs import make_compression_codec
from .codecs import make_compression_lazy_loaded_codec


if ta.TYPE_CHECKING:
    import gzip
    import zlib
else:
    gzip = lang.proxy_import('gzip')
    zlib = lang.proxy_import('zlib')


##


COMPRESS_LEVEL_FAST = 1
COMPRESS_LEVEL_TRADEOFF = 6
COMPRESS_LEVEL_BEST = 9


@dc.dataclass(frozen=True, kw_only=True)
class GzipCompression(Compression, IncrementalCompression):
    level: int = COMPRESS_LEVEL_BEST

    mtime: float | None = None

    def compress(self, d: bytes) -> bytes:
        return gzip.compress(
            d,
            self.level,
            mtime=self.mtime,
        )

    def decompress(self, d: bytes) -> bytes:
        return gzip.decompress(
            d,
        )

    def compress_incremental(self) -> BytesSteppedGenerator[None]:
        return lang.nextgen(IncrementalGzipCompressor(
            level=self.level,
            mtime=self.mtime,
        )())

    def decompress_incremental(self) -> BytesSteppedReaderGenerator[None]:
        return IncrementalGzipDecompressor()()


##


@cached.function
def _zero_crc() -> int:
    return zlib.crc32(b'')


##


class IncrementalGzipCompressor:
    def __init__(
            self,
            *,
            level: int = COMPRESS_LEVEL_BEST,
            name: str | bytes | None = None,
            mtime: float | None = None,
    ) -> None:
        super().__init__()

        self._name = name or ''
        self._level = level
        self._mtime = mtime

    def _write_gzip_header(self) -> ta.Generator[bytes, None, None]:
        check.none((yield b'\037\213'))  # magic header
        check.none((yield b'\010'))  # compression method

        try:
            # RFC 1952 requires the FNAME field to be Latin-1. Do not include filenames that cannot be represented that
            # way.
            fname = os.path.basename(self._name)
            if not isinstance(fname, bytes):
                fname = fname.encode('latin-1')
            if fname.endswith(b'.gz'):
                fname = fname[:-3]
        except UnicodeEncodeError:
            fname = b''

        flags = 0
        if fname:
            flags = gzip.FNAME
        check.none((yield chr(flags).encode('latin-1')))

        mtime = self._mtime
        if mtime is None:
            mtime = time.time()
        check.none((yield struct.pack('<L', int(mtime))))

        if self._level == COMPRESS_LEVEL_BEST:
            xfl = b'\002'
        elif self._level == COMPRESS_LEVEL_FAST:
            xfl = b'\004'
        else:
            xfl = b'\000'
        check.none((yield xfl))

        check.none((yield b'\377'))

        if fname:
            check.none((yield fname + b'\000'))

    def __call__(self) -> BytesSteppedGenerator:
        crc = _zero_crc()
        size = 0
        offset = 0  # Current file offset for seek(), tell(), etc
        wrote_header = False

        compress = zlib.compressobj(
            self._level,
            zlib.DEFLATED,
            -zlib.MAX_WBITS,
            zlib.DEF_MEM_LEVEL,
            0,
        )

        while True:
            data: ta.Any = check.isinstance((yield None), bytes)

            if not wrote_header:
                yield from self._write_gzip_header()
                wrote_header = True

            if not data:
                break

            # Called by our self._buffer underlying BufferedWriterDelegate.
            if isinstance(data, (bytes, bytearray)):
                length = len(data)
            else:
                # accept any data that supports the buffer protocol
                data = memoryview(data)
                length = data.nbytes

            if length > 0:
                if (fl := compress.compress(data)):
                    check.none((yield fl))
                size += length
                crc = zlib.crc32(data, crc)
                offset += length

        if (fl := compress.flush()):
            check.none((yield fl))

        yield struct.pack('<L', crc)
        # size may exceed 2 GiB, or even 4 GiB
        yield struct.pack('<L', size & 0xffffffff)

        yield b''


##


class IncrementalGzipDecompressor:
    def __init__(self) -> None:
        super().__init__()

        self._factory = functools.partial(
            zlib.decompressobj,
            wbits=-zlib.MAX_WBITS,
        )

    def _read_gzip_header(
            self,
            rdr: PrependableBytesGeneratorReader,
    ) -> ta.Generator[int | None, bytes, int | None]:
        magic = yield from rdr.read(2)
        if magic == b'':
            return None

        if magic != b'\037\213':
            raise gzip.BadGzipFile(f'Not a gzipped file ({magic!r})')

        buf = yield from rdr.read(8)
        method, flag, last_mtime = struct.unpack('<BBIxx', buf)
        if method != 8:
            raise gzip.BadGzipFile('Unknown compression method')

        if flag & gzip.FEXTRA:
            # Read & discard the extra field, if present
            buf = yield from rdr.read(2)
            extra_len, = struct.unpack('<H', buf)
            if extra_len:
                yield from rdr.read(extra_len)

        if flag & gzip.FNAME:
            # Read and discard a null-terminated string containing the filename
            while True:
                s = yield from rdr.read(1)
                if not s or s == b'\000':
                    break

        if flag & gzip.FCOMMENT:
            # Read and discard a null-terminated string containing a comment
            while True:
                s = yield from rdr.read(1)
                if not s or s == b'\000':
                    break

        if flag & gzip.FHCRC:
            yield from rdr.read(2)  # Read & discard the 16-bit header CRC

        return last_mtime

    def _read_eof(
            self,
            rdr: PrependableBytesGeneratorReader,
            crc: int,
            stream_size: int,
    ) -> ta.Generator[int | None, bytes, None]:
        # We've read to the end of the file.
        # We check that the computed CRC and size of the uncompressed data matches the stored values. Note that the size
        # stored is the true file size mod 2**32.
        buf = yield from rdr.read(8)
        crc32, isize = struct.unpack('<II', buf)
        if crc32 != crc:
            raise gzip.BadGzipFile(f'CRC check failed {hex(crc32)} != {hex(crc)}')
        elif isize != (stream_size & 0xffffffff):
            raise gzip.BadGzipFile('Incorrect length of data produced')

        # Gzip files can be padded with zeroes and still have archives. Consume all zero bytes and set the file position
        # to the first non-zero byte. See http://www.gzip.org/#faq8
        c = b'\0'
        while c == b'\0':
            c = yield from rdr.read(1)
        if c:
            rdr.prepend(c)

    def __call__(self) -> BytesSteppedReaderGenerator:
        rdr = PrependableBytesGeneratorReader()

        pos = 0  # Current offset in decompressed stream

        crc = _zero_crc()
        stream_size = 0  # Decompressed size of unconcatenated stream
        new_member = True

        decompressor = self._factory()

        while True:
            # For certain input data, a single call to decompress() may not return any data. In this case, retry until
            # we get some data or reach EOF.
            while True:
                if decompressor.eof:
                    # Ending case: we've come to the end of a member in the file, so finish up this member, and read a
                    # new gzip header. Check the CRC and file size, and set the flag so we read a new member
                    yield from self._read_eof(rdr, crc, stream_size)
                    new_member = True
                    decompressor = self._factory()

                if new_member:
                    # If the _new_member flag is set, we have to jump to the next member, if there is one.
                    crc = _zero_crc()
                    stream_size = 0  # Decompressed size of unconcatenated stream
                    last_mtime = yield from self._read_gzip_header(rdr)
                    if not last_mtime:
                        check.none((yield b''))
                        return
                    new_member = False

                # Read a chunk of data from the file
                if not decompressor.unconsumed_tail:
                    buf = yield from rdr.read(None)
                    uncompress = decompressor.decompress(buf)
                else:
                    uncompress = decompressor.decompress(b'')

                if decompressor.unused_data != b'':
                    # Prepend the already read bytes to the fileobj so they can be seen by _read_eof() and
                    # _read_gzip_header()
                    rdr.prepend(decompressor.unused_data)

                if uncompress != b'':
                    break
                if buf == b'':  # noqa
                    raise EOFError('Compressed file ended before the end-of-stream marker was reached')

            crc = zlib.crc32(uncompress, crc)
            stream_size += len(uncompress)
            pos += len(uncompress)
            check.none((yield uncompress))


##


GZIP_CODEC = make_compression_codec('gzip', GzipCompression, aliases=['gz'])

# @omlish-manifest
_GZIP_LAZY_CODEC = make_compression_lazy_loaded_codec(__name__, 'GZIP_CODEC', GZIP_CODEC)



import abc
import typing as ta
import weakref

from .dispatch import find_impl


T = ta.TypeVar('T')


##


class DispatchCacheProtocol(ta.Protocol[T]):
    def size(self) -> int: ...
    def prepare(self, cls: type) -> None: ...
    def clear(self) -> None: ...
    def put(self, cls: type, impl: T) -> None: ...
    def get(self, cls: type) -> T: ...  # Raises[KeyError]


class DispatcherProtocol(ta.Protocol[T]):
    def cache_size(self) -> int: ...
    def register(self, impl: T, cls_col: ta.Iterable[type]) -> T: ...
    def dispatch(self, cls: type) -> T | None: ...


##


class DispatchCache(DispatchCacheProtocol[T]):
    def __init__(self) -> None:
        super().__init__()

        self._dct: dict[ta.Any, T] = {}

        def remove(k, self_ref=weakref.ref(self)):
            if (ref_self := self_ref()) is not None:
                dct = ref_self._dct  # noqa
                try:
                    del dct[k]
                except KeyError:
                    pass

        self._remove = remove

        self._token: ta.Any = None

    def size(self) -> int:
        return len(self._dct)

    def prepare(self, cls: type) -> None:
        if self._token is None and hasattr(cls, '__abstractmethods__'):
            self._token = abc.get_cache_token()

        self.clear()

    def clear(self) -> None:
        if self._dct:
            self._dct.clear()

    def put(self, cls: type, impl: T) -> None:
        self._dct[weakref.ref(cls, self._remove)] = impl

    def get(self, cls: type) -> T:
        if self._token is not None and (current_token := abc.get_cache_token()) != self._token:
            self._dct.clear()
            self._token = current_token

        cls_ref = weakref.ref(cls)
        return self._dct[cls_ref]


class Dispatcher(DispatcherProtocol[T]):
    def __init__(self) -> None:
        super().__init__()

        self._impls_by_arg_cls: dict[type, T] = {}
        self._cache: DispatchCache[T | None] = DispatchCache()

    def cache_size(self) -> int:
        return self._cache.size()

    def register(self, impl: T, cls_col: ta.Iterable[type]) -> T:
        for cls in cls_col:
            self._impls_by_arg_cls[cls] = impl
            self._cache.prepare(cls)

        return impl

    def dispatch(self, cls: type) -> T | None:
        try:
            return self._cache.get(cls)
        except KeyError:
            pass

        impl: T | None
        try:
            impl = self._impls_by_arg_cls[cls]
        except KeyError:
            impl = find_impl(cls, self._impls_by_arg_cls)

        self._cache.put(cls, impl)
        return impl



# type: ignore
# ruff: noqa
# flake8: noqa
# Copyright (c) 2012-2017 The ANTLR Project. All rights reserved.
# Use of this file is governed by the BSD 3-clause license that
# can be found in the LICENSE.txt file in the project root.
#

from io import StringIO

def str_list(val):
    with StringIO() as buf:
        buf.write('[')
        first = True
        for item in val:
            if not first:
                buf.write(', ')
            buf.write(str(item))
            first = False
        buf.write(']')
        return buf.getvalue()

def escapeWhitespace(s:str, escapeSpaces:bool):
    with StringIO() as buf:
        for c in s:
            if c==' ' and escapeSpaces:
                buf.write('\u00B7')
            elif c=='\t':
                buf.write("\\t")
            elif c=='\n':
                buf.write("\\n")
            elif c=='\r':
                buf.write("\\r")
            else:
                buf.write(c)
        return buf.getvalue()



import inspect
import typing as ta

from .. import check
from .. import collections as col
from .. import lang
from .. import reflect as rfl
from .base import MarshalContext
from .base import Marshaler
from .base import MarshalerFactory
from .base import Option
from .base import UnmarshalContext
from .base import Unmarshaler
from .base import UnmarshalerFactory
from .objects import FieldInfo
from .objects import FieldInfos
from .objects import FieldMetadata
from .objects import ObjectMarshaler
from .objects import ObjectUnmarshaler


##


def _is_named_tuple(rty: rfl.Type) -> bool:
    return (
        isinstance(rty, type) and
        issubclass(rty, tuple) and
        ta.NamedTuple in rfl.get_orig_bases(rty)
    )


def get_field_infos(
        ty: type,
        opts: col.TypeMap[Option] = col.TypeMap(),
) -> FieldInfos:
    check.arg(_is_named_tuple(ty), ty)

    sig = inspect.signature(ty)

    ret: list[FieldInfo] = []
    for param in sig.parameters.values():
        ret.append(FieldInfo(
            name=param.name,
            type=param.annotation,
            metadata=FieldMetadata(),

            marshal_name=param.name,
            unmarshal_names=[param.name],
        ))

    return FieldInfos(ret)


##


class NamedtupleMarshalerFactory(MarshalerFactory):
    def guard(self, ctx: MarshalContext, rty: rfl.Type) -> bool:
        return _is_named_tuple(rty)

    def fn(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        check.state(_is_named_tuple(rty))
        ty = check.isinstance(rty, type)
        check.state(not lang.is_abstract_class(ty))

        fis = get_field_infos(ty, ctx.options)

        fields = [
            (fi, ctx.make(fi.type))
            for fi in fis
        ]

        return ObjectMarshaler(
            fields,
        )


##


class NamedtupleUnmarshalerFactory(UnmarshalerFactory):
    def guard(self, ctx: UnmarshalContext, rty: rfl.Type) -> bool:
        return _is_named_tuple(rty)

    def fn(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        check.state(_is_named_tuple(rty))
        ty = check.isinstance(rty, type)
        check.state(not lang.is_abstract_class(ty))

        fis = get_field_infos(ty, ctx.options)

        d: dict[str, tuple[FieldInfo, Unmarshaler]] = {}
        defaults: dict[str, ta.Any] = {}

        for fi in fis:
            tup = (fi, ctx.make(fi.type))

            for un in fi.unmarshal_names:
                if un in d:
                    raise KeyError(f'Duplicate fields for name {un!r}: {fi.name!r}, {d[un][0].name!r}')
                d[un] = tup

            if fi.options.default.present:
                defaults[fi.name] = fi.options.default.must()

        return ObjectUnmarshaler(
            ty,
            d,
            defaults=defaults,
        )



"""
https://github.com/pytest-dev/pytest/blob/72c682ff9773ad2690711105a100423ebf7c7c15/src/_pytest/python.py#L494

--

https://github.com/pytest-dev/pytest-asyncio/blob/b1dc0c3e2e82750bdc6dbdf668d519aaa89c036c/pytest_asyncio/plugin.py#L657
"""
import dataclasses as dc
import re
import typing as ta

import pytest

from .... import check
from ._registry import register as register_plugin
from .utils import find_plugin


@register_plugin
class DepSkipPlugin:
    @dc.dataclass(frozen=True)
    class Entry:
        file_pats: ta.Sequence[re.Pattern]
        imp_pats: ta.Sequence[re.Pattern]

    def __init__(self) -> None:
        super().__init__()

        self._entries: list[DepSkipPlugin.Entry] = []

    def add_entry(self, e: Entry) -> None:
        self._entries.append(e)

    def should_skip(self, file_name: str, imp_name: str) -> bool:
        for e in self._entries:
            if (
                any(fp.fullmatch(file_name) for fp in e.file_pats) and
                any(ip.fullmatch(imp_name) for ip in e.imp_pats)
            ):
                return True
        return False

    @pytest.hookimpl
    def pytest_collectstart(self, collector: pytest.Collector) -> None:
        if isinstance(collector, pytest.Module):
            original_attr = f'__{self.__class__.__qualname__.replace(".", "__")}__original_getobj'

            def _patched_getobj():
                try:
                    return getattr(collector, original_attr)()
                except pytest.Collector.CollectError as ce:
                    if (oe := ce.__cause__) and isinstance(oe, ImportError):
                        if (
                                (file_name := collector.nodeid) and
                                (imp_name := oe.name) and
                                self.should_skip(file_name, imp_name)  # noqa
                        ):
                            pytest.skip(
                                f'skipping {file_name} to missing optional dependency {imp_name}',
                                allow_module_level=True,
                            )

                    raise

            setattr(collector, original_attr, collector._getobj)  # noqa
            collector._getobj = _patched_getobj  # type: ignore  # noqa


def register(
        pm: pytest.PytestPluginManager,
        file_pats: ta.Iterable[str],
        imp_pats: ta.Iterable[str],
) -> None:
    check.not_isinstance(file_pats, str)
    check.not_isinstance(imp_pats, str)

    pg = check.not_none(find_plugin(pm, DepSkipPlugin))
    pg.add_entry(DepSkipPlugin.Entry(
        [re.compile(fp) for fp in file_pats],
        [re.compile(ip) for ip in imp_pats],
    ))



import contextlib
import importlib
import threading
import typing as ta

from .. import cached
from .. import check
from .. import lang
from .base import Codec
from .base import LazyLoadedCodec
from .standard import STANDARD_CODECS


if ta.TYPE_CHECKING:
    from ..manifests import load as manifest_load
else:
    manifest_load = lang.proxy_import('..manifests.load', __package__)


##


class CodecRegistry:
    def __init__(
            self,
            *,
            late_load_callbacks: ta.Iterable[ta.Callable[['CodecRegistry'], None]] | None = None,
    ) -> None:
        super().__init__()

        self._late_load_callbacks = late_load_callbacks

        self._lock = threading.RLock()
        self._by_name: dict[str, Codec | LazyLoadedCodec] = {}
        self._names_by_alias: dict[str, str] = {}
        self._names_by_cls: dict[type, list[str]] = {}

    def _late_load(self) -> None:
        if self._late_load_callbacks:
            for cb in self._late_load_callbacks:
                cb(self)
            self._late_load_callbacks = None

    @contextlib.contextmanager
    def _lock_and_load(self) -> ta.Iterator[None]:
        with self._lock:
            self._late_load()
            yield

    def _post_load(self, codec: Codec) -> None:
        for t in type(codec).__mro__:
            if t is not object:
                self._names_by_cls.setdefault(t, []).append(codec.name)

    def register(self, *codecs: Codec | LazyLoadedCodec) -> ta.Self:
        with self._lock:
            for codec in codecs:
                for n in {codec.name, *(codec.aliases or [])}:
                    if n in self._names_by_alias:
                        raise KeyError(n)

            for codec in codecs:
                self._by_name[codec.name] = codec
                for n in {codec.name, *(codec.aliases or [])}:
                    self._names_by_alias[n] = codec.name
                if isinstance(codec, Codec):
                    self._post_load(codec)

        return self

    def lookup(self, name_or_alias: str) -> Codec:
        with self._lock_and_load():
            name = self._names_by_alias[name_or_alias.replace('_', '-')]
            codec_or_lazy = self._by_name[name]

            if isinstance(codec_or_lazy, LazyLoadedCodec):
                mod = importlib.import_module(codec_or_lazy.mod_name)
                codec = check.isinstance(getattr(mod, codec_or_lazy.attr_name), Codec)
                self._by_name[name] = codec
                self._post_load(codec)
            else:
                codec = check.isinstance(codec_or_lazy, Codec)

            return codec

    def lookup_type(self, cls: type) -> list[Codec]:
        with self._lock_and_load():
            return [self.lookup(n) for n in self._names_by_cls.get(cls, [])]

    def all(self) -> frozenset[str]:
        with self._lock_and_load():
            return frozenset(self._by_name)


##


def _install_standard_codecs(registry: CodecRegistry) -> None:
    registry.register(*STANDARD_CODECS)


##


@cached.function
def _build_manifest_lazy_loaded_codecs() -> ta.Sequence[LazyLoadedCodec]:
    ldr = manifest_load.MANIFEST_LOADER
    pkgs = {__package__.split('.')[0], *ldr.discover()}
    mns = ldr.load(*pkgs, only=[LazyLoadedCodec])
    return [m.value for m in mns]


def _install_manifest_lazy_loaded_codecs(registry: CodecRegistry) -> None:
    registry.register(*_build_manifest_lazy_loaded_codecs())


##


REGISTRY = CodecRegistry(
    late_load_callbacks=[
        _install_standard_codecs,
        _install_manifest_lazy_loaded_codecs,
    ],
)

register = REGISTRY.register
lookup = REGISTRY.lookup


##


def encode(i: ta.Any, name: str, **kwargs: ta.Any) -> ta.Any:
    return lookup(name).new(**kwargs).encode(i)


def decode(o: ta.Any, name: str, **kwargs: ta.Any) -> ta.Any:
    return lookup(name).new(**kwargs).decode(o)



"""
TODO:
 - cfg naming
 - adapters for dataclasses / namedtuples / user objects (as configured)
 - mro-merge ObjectMetadata
 - key ordering override - like slice, -1 means last
"""
import collections.abc
import typing as ta

from .. import cached
from .. import check
from .. import collections as col
from .. import dataclasses as dc
from .. import lang
from .. import reflect as rfl
from .base import MarshalContext
from .base import Marshaler
from .base import MarshalerFactory
from .base import UnmarshalContext
from .base import Unmarshaler
from .base import UnmarshalerFactory
from .naming import Naming
from .values import Value


##


@dc.dataclass(frozen=True, kw_only=True)
class FieldOptions:
    omit_if: ta.Callable[[ta.Any], bool] | None = None

    default: lang.Maybe[ta.Any] = dc.xfield(default=lang.empty(), check_type=lang.Maybe)

    embed: bool = False

    no_marshal: bool = False
    no_unmarshal: bool = False


DEFAULT_FIELD_OPTIONS = FieldOptions()
FIELD_OPTIONS_KWARGS: frozenset[str] = frozenset(dc.fields_dict(FieldOptions).keys())


@dc.dataclass(frozen=True, kw_only=True)
class FieldMetadata:
    name: str | None = None
    alts: ta.Iterable[str] | None = None

    options: FieldOptions = DEFAULT_FIELD_OPTIONS

    marshaler: Marshaler | None = dc.xfield(None, check_type=(Marshaler, None))
    marshaler_factory: MarshalerFactory | None = None

    unmarshaler: Unmarshaler | None = dc.xfield(None, check_type=(Unmarshaler, None))
    unmarshaler_factory: UnmarshalerFactory | None = None

    def update(self, **kwargs: ta.Any) -> 'FieldMetadata':
        okw = {k: v for k, v in kwargs.items() if k in FIELD_OPTIONS_KWARGS}
        mkw = {k: v for k, v in kwargs.items() if k not in FIELD_OPTIONS_KWARGS}
        return dc.replace(
            self,
            **(dict(options=dc.replace(self.options, **okw)) if okw else {}),
            **mkw,
        )


@dc.dataclass(frozen=True, kw_only=True)
class ObjectMetadata:
    field_naming: Naming | None = None

    unknown_field: str | None = None
    source_field: str | None = None

    @cached.property
    def specials(self) -> 'ObjectSpecials':
        return ObjectSpecials(
            unknown=self.unknown_field,
            source=self.source_field,
        )

    field_defaults: FieldMetadata = FieldMetadata()

    ignore_unknown: bool = False


@dc.dataclass(frozen=True, kw_only=True)
class ObjectSpecials:
    unknown: str | None = None
    source: str | None = None

    @cached.property
    def set(self) -> frozenset[str]:
        return frozenset(v for v in dc.asdict(self).values() if v is not None)


##


@dc.dataclass(frozen=True, kw_only=True)
class FieldInfo:
    name: str
    type: ta.Any

    marshal_name: str | None
    unmarshal_names: ta.Sequence[str]

    metadata: FieldMetadata = FieldMetadata()

    options: FieldOptions = FieldOptions()


@dc.dataclass(frozen=True)
class FieldInfos:
    lst: ta.Sequence[FieldInfo]

    def __iter__(self) -> ta.Iterator[FieldInfo]:
        return iter(self.lst)

    def __len__(self) -> int:
        return len(self.lst)

    @cached.property
    @dc.init
    def by_name(self) -> ta.Mapping[str, FieldInfo]:
        return col.make_map(((fi.name, fi) for fi in self), strict=True)

    @cached.property
    @dc.init
    def by_marshal_name(self) -> ta.Mapping[str, FieldInfo]:
        return col.make_map(((fi.marshal_name, fi) for fi in self if fi.marshal_name is not None), strict=True)

    @cached.property
    @dc.init
    def by_unmarshal_name(self) -> ta.Mapping[str, FieldInfo]:
        return col.make_map(((n, fi) for fi in self for n in fi.unmarshal_names), strict=True)


##


@dc.dataclass(frozen=True)
class ObjectMarshaler(Marshaler):
    fields: ta.Sequence[tuple[FieldInfo, Marshaler]]

    _: dc.KW_ONLY

    specials: ObjectSpecials = ObjectSpecials()

    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        ret: dict[str, ta.Any] = {}
        for fi, m in self.fields:
            v = getattr(o, fi.name)

            if fi.options.omit_if is not None and fi.options.omit_if(v):
                continue

            if fi.name in self.specials.set:
                continue

            mn = fi.marshal_name
            if mn is None:
                continue

            mv = m.marshal(ctx, v)

            if fi.options.embed:
                for ek, ev in check.isinstance(mv, collections.abc.Mapping).items():
                    ret[mn + check.non_empty_str(ek)] = ev

            else:
                ret[mn] = mv

        if self.specials.unknown is not None:
            if (ukf := getattr(o, self.specials.unknown)):
                if (dks := set(ret) & set(ukf)):
                    raise KeyError(f'Unknown field keys duplicate fields: {dks!r}')

            ret.update(ukf)  # FIXME: marshal?

        return ret


@dc.dataclass(frozen=True)
class SimpleObjectMarshalerFactory(MarshalerFactory):
    dct: ta.Mapping[type, ta.Sequence[FieldInfo]]

    _: dc.KW_ONLY

    specials: ObjectSpecials = ObjectSpecials()

    def guard(self, ctx: MarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, type) and rty in self.dct

    def fn(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        ty = check.isinstance(rty, type)

        flds = FieldInfos(self.dct[ty])

        fields = [
            (fi, ctx.make(fi.type))
            for fi in flds
        ]

        return ObjectMarshaler(
            fields,
            specials=self.specials,
        )


##


@dc.dataclass(frozen=True)
class ObjectUnmarshaler(Unmarshaler):
    cls: type
    fields_by_unmarshal_name: ta.Mapping[str, tuple[FieldInfo, Unmarshaler]]

    _: dc.KW_ONLY

    specials: ObjectSpecials = ObjectSpecials()

    defaults: ta.Mapping[str, ta.Any] | None = None

    embeds: ta.Mapping[str, type] | None = None
    embeds_by_unmarshal_name: ta.Mapping[str, tuple[str, str]] | None = None

    ignore_unknown: bool = False

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        ma = check.isinstance(v, collections.abc.Mapping)

        u: ta.Any
        kw: dict[str, ta.Any] = {}
        ukf: dict[str, ta.Any] | None = None

        ekws: dict[str, dict[str, ta.Any]] = {en: {} for en in self.embeds or ()}

        if self.specials.source is not None:
            kw[self.specials.source] = v

        if self.specials.unknown is not None:
            kw[self.specials.unknown] = ukf = {}

        for k, mv in ma.items():
            ks = check.isinstance(k, str)

            try:
                fi, u = self.fields_by_unmarshal_name[ks]

            except KeyError:
                if ukf is not None:
                    ukf[ks] = mv  # FIXME: unmarshal?
                    continue

                if self.ignore_unknown:
                    continue

                raise

            if self.embeds_by_unmarshal_name and (en := self.embeds_by_unmarshal_name.get(ks)):
                tkw, tk = ekws[en[0]], en[1]
            else:
                tkw, tk = kw, fi.name

            if tk in tkw:
                raise KeyError(f'Duplicate keys for field {tk!r}: {ks!r}')

            tkw[tk] = u.unmarshal(ctx, mv)

        for em, ecls in self.embeds.items() if self.embeds else ():
            ekw = ekws[em]
            ev = ecls(**ekw)
            kw[em] = ev

        if self.defaults:
            for dk, dv in self.defaults.items():
                kw.setdefault(dk, dv)

        return self.cls(**kw)


@dc.dataclass(frozen=True)
class SimpleObjectUnmarshalerFactory(UnmarshalerFactory):
    dct: ta.Mapping[type, ta.Sequence[FieldInfo]]

    _: dc.KW_ONLY

    specials: ObjectSpecials = ObjectSpecials()

    def guard(self, ctx: UnmarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, type) and rty in self.dct

    def fn(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        ty = check.isinstance(rty, type)

        flds = FieldInfos(self.dct[ty])

        fields_by_unmarshal_name = {
            n: (fi, ctx.make(fi.type))
            for fi in flds
            for n in fi.unmarshal_names
        }

        defaults = {
            fi.name: fi.options.default.must()
            for fi in flds
            if fi.options.default.present
        }

        return ObjectUnmarshaler(
            ty,
            fields_by_unmarshal_name,
            specials=self.specials,
            defaults=defaults,
        )



# type: ignore
# ruff: noqa
# flake8: noqa
#
# Copyright (c) 2012-2017 The ANTLR Project. All rights reserved.
# Use of this file is governed by the BSD 3-clause license that
# can be found in the LICENSE.txt file in the project root.

from ..PredictionContext import merge
from ..Utils import str_list
from .ATN import ATN
from .ATNConfig import ATNConfig
from .SemanticContext import SemanticContext
from ..error.Errors import UnsupportedOperationException, IllegalStateException
from functools import reduce
#
# Specialized {@link Set}{@code <}{@link ATNConfig}{@code >} that can track
# info about the set, with support for combining similar configurations using a
# graph-structured stack.
# /
from io import StringIO

ATNSimulator = None

class ATNConfigSet(object):
    __slots__ = (
        'configLookup', 'fullCtx', 'readonly', 'configs', 'uniqueAlt',
        'conflictingAlts', 'hasSemanticContext', 'dipsIntoOuterContext',
        'cachedHashCode'
    )

    #
    # The reason that we need this is because we don't want the hash map to use
    # the standard hash code and equals. We need all configurations with the same
    # {@code (s,i,_,semctx)} to be equal. Unfortunately, this key effectively doubles
    # the number of objects associated with ATNConfigs. The other solution is to
    # use a hash table that lets us specify the equals/hashcode operation.

    def __init__(self, fullCtx:bool=True):
        # All configs but hashed by (s, i, _, pi) not including context. Wiped out
        # when we go readonly as this set becomes a DFA state.
        self.configLookup = dict()
        # Indicates that this configuration set is part of a full context
        #  LL prediction. It will be used to determine how to merge $. With SLL
        #  it's a wildcard whereas it is not for LL context merge.
        self.fullCtx = fullCtx
        # Indicates that the set of configurations is read-only. Do not
        #  allow any code to manipulate the set; DFA states will point at
        #  the sets and they must not change. This does not protect the other
        #  fields; in particular, conflictingAlts is set after
        #  we've made this readonly.
        self.readonly = False
        # Track the elements as they are added to the set; supports get(i)#/
        self.configs = []

        # TODO: these fields make me pretty uncomfortable but nice to pack up info together, saves recomputation
        # TODO: can we track conflicts as they are added to save scanning configs later?
        self.uniqueAlt = 0
        self.conflictingAlts = None

        # Used in parser and lexer. In lexer, it indicates we hit a pred
        # while computing a closure operation.  Don't make a DFA state from this.
        self.hasSemanticContext = False
        self.dipsIntoOuterContext = False

        self.cachedHashCode = -1

    def __iter__(self):
        return self.configs.__iter__()

    # Adding a new config means merging contexts with existing configs for
    # {@code (s, i, pi, _)}, where {@code s} is the
    # {@link ATNConfig#state}, {@code i} is the {@link ATNConfig#alt}, and
    # {@code pi} is the {@link ATNConfig#semanticContext}. We use
    # {@code (s,i,pi)} as key.
    #
    # <p>This method updates {@link #dipsIntoOuterContext} and
    # {@link #hasSemanticContext} when necessary.</p>
    #/
    def add(self, config:ATNConfig, mergeCache=None):
        if self.readonly:
            raise Exception("This set is readonly")
        if config.semanticContext is not SemanticContext.NONE:
            self.hasSemanticContext = True
        if config.reachesIntoOuterContext > 0:
            self.dipsIntoOuterContext = True
        existing = self.getOrAdd(config)
        if existing is config:
            self.cachedHashCode = -1
            self.configs.append(config)  # track order here
            return True
        # a previous (s,i,pi,_), merge with it and save result
        rootIsWildcard = not self.fullCtx
        merged = merge(existing.context, config.context, rootIsWildcard, mergeCache)
        # no need to check for existing.context, config.context in cache
        # since only way to create new graphs is "call rule" and here.
        # We cache at both places.
        existing.reachesIntoOuterContext = max(existing.reachesIntoOuterContext, config.reachesIntoOuterContext)
        # make sure to preserve the precedence filter suppression during the merge
        if config.precedenceFilterSuppressed:
            existing.precedenceFilterSuppressed = True
        existing.context = merged # replace context; no need to alt mapping
        return True

    def getOrAdd(self, config:ATNConfig):
        h = config.hashCodeForConfigSet()
        l = self.configLookup.get(h, None)
        if l is not None:
            r = next((cfg for cfg in l if config.equalsForConfigSet(cfg)), None)
            if r is not None:
                return r
        if l is None:
            l = [config]
            self.configLookup[h] = l
        else:
            l.append(config)
        return config

    def getStates(self):
        return set(c.state for c in self.configs)

    def getPredicates(self):
        return list(cfg.semanticContext for cfg in self.configs if cfg.semanticContext!=SemanticContext.NONE)

    def get(self, i:int):
        return self.configs[i]

    def optimizeConfigs(self, interpreter:ATNSimulator):
        if self.readonly:
            raise IllegalStateException("This set is readonly")
        if len(self.configs)==0:
            return
        for config in self.configs:
            config.context = interpreter.getCachedContext(config.context)

    def addAll(self, coll:list):
        for c in coll:
            self.add(c)
        return False

    def __eq__(self, other):
        if self is other:
            return True
        elif not isinstance(other, ATNConfigSet):
            return False

        same = self.configs is not None and \
            self.configs==other.configs and \
            self.fullCtx == other.fullCtx and \
            self.uniqueAlt == other.uniqueAlt and \
            self.conflictingAlts == other.conflictingAlts and \
            self.hasSemanticContext == other.hasSemanticContext and \
            self.dipsIntoOuterContext == other.dipsIntoOuterContext

        return same

    def __hash__(self):
        if self.readonly:
            if self.cachedHashCode == -1:
                self.cachedHashCode = self.hashConfigs()
            return self.cachedHashCode
        return self.hashConfigs()

    def hashConfigs(self):
        return reduce(lambda h, cfg: hash((h, cfg)), self.configs, 0)

    def __len__(self):
        return len(self.configs)

    def isEmpty(self):
        return len(self.configs)==0

    def __contains__(self, config):
        if self.configLookup is None:
            raise UnsupportedOperationException("This method is not implemented for readonly sets.")
        h = config.hashCodeForConfigSet()
        l = self.configLookup.get(h, None)
        if l is not None:
            for c in l:
                if config.equalsForConfigSet(c):
                    return True
        return False

    def clear(self):
        if self.readonly:
            raise IllegalStateException("This set is readonly")
        self.configs.clear()
        self.cachedHashCode = -1
        self.configLookup.clear()

    def setReadonly(self, readonly:bool):
        self.readonly = readonly
        self.configLookup = None # can't mod, no need for lookup cache

    def __str__(self):
        with StringIO() as buf:
            buf.write(str_list(self.configs))
            if self.hasSemanticContext:
                buf.write(",hasSemanticContext=")
                buf.write(str(self.hasSemanticContext).lower()) # lower() to conform to java output
            if self.uniqueAlt!=ATN.INVALID_ALT_NUMBER:
                buf.write(",uniqueAlt=")
                buf.write(str(self.uniqueAlt))
            if self.conflictingAlts is not None:
                buf.write(",conflictingAlts=")
                buf.write(str(self.conflictingAlts))
            if self.dipsIntoOuterContext:
                buf.write(",dipsIntoOuterContext")
            return buf.getvalue()


class OrderedATNConfigSet(ATNConfigSet):

    def __init__(self):
        super().__init__()



# ruff: noqa: ANN204

class IOBase:
    def seek(self, pos, whence=0): ...

    def tell(self): ...

    def truncate(self, pos=None): ...

    def flush(self): ...

    def close(self): ...

    def seekable(self): ...

    def readable(self): ...

    def writable(self): ...

    @property
    def closed(self): ...  # noqa

    def __enter__(self): ...

    def __exit__(self, *args): ...

    def fileno(self): ...

    def isatty(self): ...

    def readline(self, size=-1): ...

    def readlines(self, hint=None): ...

    def writelines(self, lines): ...


class RawIOBase(IOBase):
    def read(self, size=-1): ...

    def readall(self): ...

    def readinto(self, b): ...

    def write(self, b): ...


class BufferedIOBase(IOBase):
    def read(self, size=-1): ...

    def read1(self, size=-1): ...

    def readinto(self, b): ...

    def readinto1(self, b): ...

    def write(self, b): ...

    def detach(self): ...


class TextIOBase(IOBase):
    def read(self, size=-1): ...

    def write(self, s): ...

    def truncate(self, pos=None): ...

    def readline(self, size=-1): ...

    def detach(self): ...

    @property
    def encoding(self): ...

    @property
    def newlines(self): ...

    @property
    def errors(self): ...



import collections.abc
import io
import typing as ta

from .. import check
from .. import collections as col
from .. import dataclasses as dc
from .. import dispatch
from .. import lang


T = ta.TypeVar('T')


Part: ta.TypeAlias = ta.Union[str, ta.Sequence['Part'], 'DataPart']
PartT = ta.TypeVar('PartT', bound=Part)


##


def _check_part(o: PartT) -> PartT:
    if isinstance(o, (str, DataPart)):
        pass
    elif isinstance(o, ta.Sequence):
        for c in o:
            _check_part(c)
    else:
        raise TypeError(o)
    return o


def _check_opt_part(o: PartT | None) -> PartT | None:
    if o is None:
        return None
    return _check_part(o)


##


class DataPart(dc.Frozen, lang.Abstract):
    pass


class Wrap(DataPart, lang.Final):
    part: Part = dc.xfield(coerce=_check_part)
    wrapper: tuple[str, str] = ('(', ')')


class List(DataPart, lang.Final):
    parts: ta.Sequence[Part | None] = dc.xfield(coerce=col.seq_of(_check_opt_part))
    delimiter: str = dc.field(default=',')  # FIXME: , check_type=str)
    trailer: bool = dc.field(default=False)  # FIXME: , check_type=bool)


class Concat(DataPart, lang.Final):
    parts: ta.Sequence[Part] = dc.xfield(coerce=col.seq_of(_check_part))


class Block(DataPart, lang.Final):
    parts: ta.Sequence[Part] = dc.xfield(coerce=col.seq_of(_check_part))


class Section(DataPart, lang.Final):
    parts: ta.Sequence[Part] = dc.xfield(coerce=col.seq_of(_check_part))


class Meta(DataPart, lang.Final):
    node: ta.Any


##


class PartTransform:
    @dispatch.method
    def __call__(self, part: Part) -> Part:
        raise TypeError(part)

    @__call__.register
    def __call__str(self, part: str) -> Part:
        return part

    @__call__.register
    def __call__sequence(self, part: collections.abc.Sequence) -> Part:
        return [self(c) for c in part]

    @__call__.register
    def __call__wrap(self, part: Wrap) -> Part:
        return Wrap(self(part.part), part.wrapper)

    @__call__.register
    def __call__list(self, part: List) -> Part:
        return List([self(c) for c in part.parts], part.delimiter, part.trailer)

    @__call__.register
    def __call__concat(self, part: Concat) -> Part:
        return Concat([self(c) for c in part.parts])

    @__call__.register
    def __call__block(self, part: Block) -> Part:
        return Block([self(c) for c in part.parts])

    @__call__.register
    def __call__section(self, part: Section) -> Part:
        return Section([self(c) for c in part.parts])

    @__call__.register
    def __call__meta(self, part: Meta) -> Meta:
        return part


##


class RemoveMetas(PartTransform):

    @PartTransform.__call__.register
    def __call__meta(self, part: Meta) -> Part:
        return []


remove_metas = RemoveMetas()


##


def _drop_empties(it: ta.Iterable[T]) -> list[T]:
    return [
        e  # type: ignore
        for e in it
        if not (
            isinstance(e, collections.abc.Sequence) and
            not e and
            not isinstance(e, str)
        )
    ]


class CompactPart(PartTransform):

    @PartTransform.__call__.register
    def __call__sequence(self, part: collections.abc.Sequence) -> Part:
        return _drop_empties(self(c) for c in part)

    @PartTransform.__call__.register
    def __call__list(self, part: List) -> Part:
        parts = _drop_empties(self(c) for c in part.parts)
        return List(parts, part.delimiter, part.trailer) if parts else []

    @PartTransform.__call__.register
    def __call__concat(self, part: Concat) -> Part:
        parts = _drop_empties(self(c) for c in part.parts)
        return Concat(parts) if parts else []

    @PartTransform.__call__.register
    def __call__block(self, part: Block) -> Part:
        parts = _drop_empties(self(c) for c in part.parts)
        return Block(parts) if parts else []

    @PartTransform.__call__.register
    def __call__section(self, part: Section) -> Part:
        parts = _drop_empties(self(c) for c in part.parts)
        return Section(parts) if parts else []


compact_part = CompactPart()


##


class PartRenderer:
    def __init__(self, buf: io.StringIO, *, indent: str = '    ') -> None:
        super().__init__()

        self._buf = buf

        self._indents = 0
        self._indent = indent

        self._blank_lines = 0
        self._has_indented = False

    def _write(self, s: str) -> None:
        check.not_in('\n', s)
        if not self._has_indented:
            self._buf.write(self._indent * self._indents)
            self._blank_lines = 0
            self._has_indented = True

        self._buf.write(s)

    def _write_newline(self, n: int = 1) -> None:
        check.arg(n >= 0)
        d = n - self._blank_lines
        if d > 0:
            self._buf.write('\n' * n)
            self._blank_lines += n
            self._has_indented = False

    @dispatch.method
    def __call__(self, part: Part) -> None:
        raise TypeError(part)

    @__call__.register
    def __call__str(self, part: str) -> None:
        self._write(part)

    @__call__.register
    def __call__sequence(self, part: collections.abc.Sequence) -> None:
        for i, c in enumerate(part):
            if i:
                self._write(' ')
            self(c)

    @__call__.register
    def __call__wrap(self, part: Wrap) -> None:
        self._write(part.wrapper[0])
        self(part.part)
        self._write(part.wrapper[1])

    @__call__.register
    def __call__list(self, part: List) -> None:
        for i, c in enumerate(part.parts):
            if i:
                self._write(part.delimiter + ' ')
            self(c)
        if part.trailer:
            self._write(part.delimiter)

    @__call__.register
    def __call__concat(self, part: Concat) -> None:
        for c in part.parts:
            self(c)

    @__call__.register
    def __call__block(self, part: Block) -> None:
        for c in part.parts:
            self(c)
            self._write_newline()

    @__call__.register
    def __call__section(self, part: Section) -> None:
        self._indents += 1
        try:
            for c in part.parts:
                self(c)
        finally:
            self._indents -= 1


def render_part(part: Part, buf: io.StringIO | None = None) -> io.StringIO:
    if buf is None:
        buf = io.StringIO()
    PartRenderer(buf)(part)
    return buf


def render(part: Part) -> str:
    part = remove_metas(part)
    part = compact_part(part)
    return render_part(part).getvalue()



import shlex


def shlex_needs_quote(s: str) -> bool:
    return bool(s) and len(list(shlex.shlex(s))) > 1


def shlex_maybe_quote(s: str) -> str:
    if shlex_needs_quote(s):
        return shlex.quote(s)
    else:
        return s



import collections.abc
import dataclasses as dc
import typing as ta

from .. import check
from .. import reflect as rfl
from ..funcs import match as mfs
from .base import MarshalContext
from .base import Marshaler
from .base import MarshalerFactoryMatchClass
from .base import UnmarshalContext
from .base import Unmarshaler
from .base import UnmarshalerFactoryMatchClass
from .values import Value


DEFAULT_MAPPING_CONCRETE_TYPES: dict[type[collections.abc.Mapping], type[collections.abc.Mapping]] = {
    collections.abc.Mapping: dict,  # type: ignore
    collections.abc.MutableMapping: dict,  # type: ignore
}


@dc.dataclass(frozen=True)
class MappingMarshaler(Marshaler):
    ke: Marshaler
    ve: Marshaler

    def marshal(self, ctx: MarshalContext, o: ta.Mapping) -> Value:
        return {
            self.ke.marshal(ctx, uk): self.ve.marshal(ctx, uv)
            for uk, uv in check.isinstance(o, collections.abc.Mapping).items()
        }


class MappingMarshalerFactory(MarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: isinstance(rty, rfl.Generic) and issubclass(rty.cls, collections.abc.Mapping))
    def _build_generic(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        gty = check.isinstance(rty, rfl.Generic)
        kt, vt = gty.args
        return MappingMarshaler(ctx.make(kt), ctx.make(vt))

    @mfs.simple(lambda _, ctx, rty: isinstance(rty, type) and issubclass(rty, collections.abc.Mapping))
    def _build_concrete(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        return MappingMarshaler(a := ctx.make(ta.Any), a)


@dc.dataclass(frozen=True)
class MappingUnmarshaler(Unmarshaler):
    ctor: ta.Callable[[ta.Mapping[ta.Any, ta.Any]], ta.Mapping]
    ke: Unmarshaler
    ve: Unmarshaler

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Mapping:
        dct: dict = {}
        for mk, mv in check.isinstance(v, collections.abc.Mapping).items():
            dct[self.ke.unmarshal(ctx, mk)] = self.ve.unmarshal(ctx, mv)
        return self.ctor(dct)


class MappingUnmarshalerFactory(UnmarshalerFactoryMatchClass):
    @mfs.simple(lambda _, ctx, rty: isinstance(rty, rfl.Generic) and issubclass(rty.cls, collections.abc.Mapping))
    def _build_generic(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        gty = check.isinstance(rty, rfl.Generic)
        cty = DEFAULT_MAPPING_CONCRETE_TYPES.get(gty.cls, gty.cls)  # noqa
        kt, vt = gty.args
        return MappingUnmarshaler(cty, ctx.make(kt), ctx.make(vt))

    @mfs.simple(lambda _, ctx, rty: isinstance(rty, type) and issubclass(rty, collections.abc.Mapping))
    def _build_concrete(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        return MappingUnmarshaler(check.isinstance(rty, type), a := ctx.make(ta.Any), a)



from ... import dataclasses as dc
from ... import lang
from ..bindings import Binding
from ..keys import Key
from ..types import Scope
from ..types import Unscoped
from .providers import ProviderImpl


##


@dc.dataclass(frozen=True, eq=False)
@dc.extra_params(cache_hash=True)
class BindingImpl(lang.Final):
    key: Key
    provider: ProviderImpl
    scope: Scope = Unscoped()
    binding: Binding | None = None
